"""Connection tests."""

import asynctest
import pytest
from pypck.connection import PchkAuthenticationError, PchkLicenseError
from pypck.lcn_addr import LcnAddr
from pypck.module import ModuleConnection


@pytest.mark.asyncio
async def test_close_without_connect(pypck_client):
    """Test closing of PchkConnectionManager without connecting."""
    await pypck_client.async_close()


@pytest.mark.asyncio
async def test_authenticate(pchk_server, pypck_client):
    """Test authentication procedure."""
    await pypck_client.async_connect()
    assert pypck_client.is_ready()


@pytest.mark.asyncio
async def test_port_error(pchk_server, pypck_client):
    """Test wrong port."""
    pypck_client.port = 55555
    with pytest.raises(ConnectionRefusedError):
        await pypck_client.async_connect()


@pytest.mark.asyncio
async def test_authentication_error(pchk_server, pypck_client):
    """Test wrong login credentials."""
    pypck_client.password = "wrong_password"
    with pytest.raises(PchkAuthenticationError):
        await pypck_client.async_connect()


@pytest.mark.asyncio
async def test_license_error(pchk_server, pypck_client):
    """Test license error."""
    pchk_server.set_license_error(True)

    with pytest.raises(PchkLicenseError):
        await pypck_client.async_connect()


@pytest.mark.asyncio
async def test_timeout_error(pchk_server, pypck_client):
    """Test timeout when connecting."""
    with pytest.raises(TimeoutError):
        await pypck_client.async_connect(timeout=0)


@pytest.mark.asyncio
async def test_lcn_connected(pchk_server, pypck_client):
    """Test lcn disconnected event."""

    pypck_client.event_handler = asynctest.CoroutineMock()
    await pypck_client.async_connect()
    await pchk_server.send_message("$io:#LCN:connected")
    await pypck_client.received("$io:#LCN:connected")

    pypck_client.event_handler.assert_has_awaits(
        [
            asynctest.mock.call("lcn-connection-status-changed"),
            asynctest.mock.call("lcn-connected"),
        ]
    )


@pytest.mark.asyncio
async def test_lcn_disconnected(pchk_server, pypck_client):
    """Test lcn disconnected event."""
    pypck_client.event_handler = asynctest.CoroutineMock()
    await pypck_client.async_connect()
    await pchk_server.send_message("$io:#LCN:disconnected")
    await pypck_client.received("$io:#LCN:disconnected")

    pypck_client.event_handler.assert_has_awaits(
        [
            asynctest.mock.call("lcn-connection-status-changed"),
            asynctest.mock.call("lcn-disconnected"),
        ]
    )


@pytest.mark.asyncio
async def test_connection_lost(pchk_server, pypck_client):
    """Test pchk server connection close."""
    pypck_client.event_handler = asynctest.CoroutineMock()
    await pypck_client.async_connect()

    await pchk_server.stop()
    # ensure that pypck_client is about to be closed
    await pypck_client.wait_closed()

    pypck_client.event_handler.assert_has_awaits(
        [asynctest.mock.call("connection-lost")]
    )


@pytest.mark.asyncio
async def test_segment_coupler_search(pchk_server, pypck_client):
    """Test segment coupler search."""
    await pypck_client.async_connect()
    await pypck_client.scan_segment_couplers(3, 0)

    assert await pchk_server.received(">G003003.SK")
    assert await pchk_server.received(">G003003.SK")
    assert await pchk_server.received(">G003003.SK")

    assert pypck_client.is_ready()


@pytest.mark.asyncio
async def test_segment_coupler_response(pchk_server, pypck_client):
    """Test segment coupler response."""
    await pypck_client.async_connect()

    assert pypck_client.local_seg_id == 0

    await pchk_server.send_message("=M000005.SK020")
    await pchk_server.send_message("=M021021.SK021")
    await pchk_server.send_message("=M022010.SK022")
    assert await pypck_client.received("=M000005.SK020")
    assert await pypck_client.received("=M021021.SK021")
    assert await pypck_client.received("=M022010.SK022")

    assert pypck_client.local_seg_id == 20
    assert set(pypck_client.segment_coupler_ids) == {20, 21, 22}


@pytest.mark.asyncio
async def test_module_scan(pchk_server, pypck_client):
    """Test module scan."""
    await pypck_client.async_connect()
    await pypck_client.scan_modules(3, 0)

    assert await pchk_server.received(">G000003!LEER")
    assert await pchk_server.received(">G000003!LEER")
    assert await pchk_server.received(">G000003!LEER")


@pytest.mark.asyncio
async def test_module_sn_response(pchk_server, pypck_client):
    """Test module scan."""
    await pypck_client.async_connect()
    module = pypck_client.get_address_conn(LcnAddr(0, 7, False))

    message = "=M000007.SN1AB20A123401FW190B11HW015"
    await pchk_server.send_message(message)
    assert await pypck_client.received(message)

    assert await module.serial_known
    assert module.hardware_serial == 0x1AB20A1234
    assert module.manu == 1
    assert module.software_serial == 0x190B11
    assert module.hw_type == 15


@pytest.mark.asyncio
async def test_send_command_to_server(pchk_server, pypck_client):
    """Test sending a command to the PCHK server."""
    await pypck_client.async_connect()
    message = ">M000007.PIN003"
    await pypck_client.async_send_command(message)
    assert await pchk_server.received(message)


@pytest.mark.asyncio
async def test_ping(pchk_server, pypck_client):
    """Test if pings are send."""
    await pypck_client.async_connect()
    assert await pchk_server.received("^ping0")


@pytest.mark.asyncio
async def test_add_address_connections(pypck_client):
    """Test if new address connections are added on request."""
    lcn_addr = LcnAddr(0, 10, False)
    assert lcn_addr not in pypck_client.address_conns

    addr_conn = pypck_client.get_address_conn(lcn_addr)
    assert isinstance(addr_conn, ModuleConnection)

    assert lcn_addr in pypck_client.address_conns


@pytest.mark.asyncio
async def test_add_address_connections_by_message(pchk_server, pypck_client):
    """Test if new address connections are added by received message."""
    await pypck_client.async_connect()
    lcn_addr = LcnAddr(0, 10, False)
    assert lcn_addr not in pypck_client.address_conns

    message = ":M000010A1050"
    await pchk_server.send_message(message)
    assert await pypck_client.received(message)

    assert lcn_addr in pypck_client.address_conns



"""Copyright (c) 2006-2018 by the respective copyright holders.

All rights reserved. This program and the accompanying materials
are made available under the terms of the Eclipse Public License v1.0
which accompanies this distribution, and is available at
http://www.eclipse.org/legal/epl-v10.html

Contributors:
  Andre Lengwenus - port to Python and further improvements
  Tobias Juettner - initial LCN binding for openHAB (Java)
"""


class LcnAddr:
    """Represents a LCN address (module or group).

    If the segment id is 0, the address object points to modules/groups which
    are in the segment where the bus coupler is connected to. This is also the
    case if no segment coupler is present at all.

    :param    int    seg_id:    Segment id
                                (0 = Local,
                                1..2 = Not allowed (but "seen in the wild")
                                3 = Broadcast,
                                4 = Status messages,
                                5..127,
                                128 = Segment-bus disabled (valid value))
    :param    bool   is_group:  Indicates whether address point to a module
                                (False) or a group (True)


    If address represents a **module**:

    :param    int    addr_id:   Module id
                                (1 = LCN-PRO,
                                2 = LCN-GVS/LCN-W,
                                4 = PCHK,
                                5..254,
                                255 = Unprog. (valid, but irrelevant here))


    If address represents a **group**:

    :param    int    addr_id:   Group id
                                (3 = Broadcast,
                                4 = Status messages,
                                5..254)
    """

    def __init__(self, seg_id: int = -1, addr_id: int = -1, is_group: bool = False):
        """Construct LcnAddr instance."""
        self.seg_id = seg_id
        self.addr_id = addr_id
        self._is_group = is_group

    def is_group(self) -> bool:
        """Get the address' module or group id (discarding the concrete type).

        :return:    Returns whether address points to a module(False) or
                    group(True)
        :rtype:     bool
        """
        return self._is_group

    def get_seg_id(self) -> int:
        """Get the logical segment id.

        :return:    The (logical) segment id
        :rtype:     int
        """
        return self.seg_id

    def get_physical_seg_id(self, local_seg_id: int) -> int:
        """Get the physical segment id ("local" segment replaced with 0).

        Can be used to send data into the LCN bus.

        :param    int    local_seg_id:    The segment id of the local segment

        :return:    The physical segment id
        :rtype:     int
        """
        return 0 if (self.seg_id == local_seg_id) else self.seg_id

    def get_id(self) -> int:
        """Get the module id.

        :return:    The module id
        :rtype:     int
        """
        return self.addr_id

    def is_valid(self) -> bool:
        """Return if the current address is valid.

        :return:    True, if address is a valid group/module address,
                    otherwise False
        :rtype:     bool
        """
        if self.is_group():
            # seg_id:
            # 0 = Local, 1..2 = Not allowed (but "seen in the wild")
            # 3 = Broadcast, 4 = Status messages, 5..127, 128 = Segment-bus
            #     disabled (valid value)
            # addr_id:
            # 3 = Broadcast, 4 = Status messages, 5..254
            is_valid = (
                (self.seg_id >= 0)
                & (self.seg_id <= 128)
                & (self.addr_id >= 3)
                & (self.addr_id <= 254)
            )
        else:
            # seg_id:
            # 0 = Local, 1..2 = Not allowed (but "seen in the wild")
            # 3 = Broadcast, 4 = Status messages, 5..127, 128 = Segment-bus
            #     disabled (valid value)
            # addr_id:
            # 1 = LCN-PRO, 2 = LCN-GVS/LCN-W, 4 = PCHK, 5..254, 255 = Unprog.
            #     (valid, but irrelevant here)
            is_valid = (
                (self.seg_id >= 0)
                & (self.seg_id <= 128)
                & (self.addr_id >= 1)
                & (self.addr_id <= 254)
            )
        return is_valid

    def __hash__(self) -> int:
        """Calculate and return hash value."""
        if self.is_valid():
            hash_value = (
                (self.is_group() << 9)
                + (reverse_uint8(self.get_id()) << 8)
                + (reverse_uint8(self.get_seg_id()))
            )
        else:
            hash_value = -1
        return hash_value

    def __eq__(self, obj: object) -> bool:
        """Return if instance equals the given object."""
        if not isinstance(obj, LcnAddr):
            return False
        return (
            (self.is_group() == obj.is_group())
            and (self.get_seg_id() == obj.get_seg_id())
            and (self.get_id() == obj.get_id())
        )


# only execute, if not defined before
if "REV_UINT8" not in dir():
    REV_UINT8 = [0] * 256
    for i in range(256):
        for j in range(8):
            if i & (1 << j) != 0:
                REV_UINT8[i] |= 0x80 >> j


def reverse_uint8(value: int) -> int:
    """Reverse the bit order of the given value."""
    if value < 0 | value > 255:
        raise ValueError("Invalid value.")
    return REV_UINT8[value]



"""Core testing functionality."""

import asyncio
from asyncio.tasks import all_tasks, wait_for

import pytest
from pypck.connection import PchkConnectionManager
from pypck.lcn_addr import LcnAddr
from pypck.pck_commands import PckGenerator

from .fake_pchk import PchkServer

HOST = "127.0.0.1"
PORT = 4114
USERNAME = "lcn_username"
PASSWORD = "lcn_password"


class MockPchkConnectionManager(PchkConnectionManager):
    def __init__(self, *args, **kwargs):
        self.data_received = []
        super().__init__(*args, **kwargs)

    async def process_message(self, message):
        await super().process_message(message)
        self.data_received.append(message)

    async def received(self, message, timeout=5, remove=True):
        async def receive_loop(data, remove):
            while data not in self.data_received:
                await asyncio.sleep(0.05)
            if remove:
                self.data_received.remove(data)

        try:
            await asyncio.wait_for(receive_loop(message, remove), timeout=timeout)
            return True
        except asyncio.TimeoutError:
            return False


def encode_pck(pck):
    """Encode the given PCK string as PCK binary string."""
    return (pck + PckGenerator.TERMINATION).encode()


@pytest.fixture
async def pchk_server():
    """Create a fake PchkServer and run."""
    ps = PchkServer(host=HOST, port=PORT, username=USERNAME, password=PASSWORD)
    await ps.run()
    yield ps
    await ps.stop()


@pytest.fixture
async def pypck_client():
    """Create a PchkConnectionManager for testing.

    Create a PchkConnection Manager for testing. Add a received coroutine method
    which returns if the specified message was received (and processed).
    """
    loop = None
    pcm = MockPchkConnectionManager(
        HOST, PORT, USERNAME, PASSWORD, settings={"SK_NUM_TRIES": 0}
    )
    yield pcm
    await pcm.async_close()


@pytest.fixture
async def module10(pypck_client):
    lcn_addr = LcnAddr(0, 10, False)
    address_connection = pypck_client.get_address_conn(lcn_addr)
    yield address_connection
    await address_connection.cancel_requests()



"""Copyright (c) 2006-2018 by the respective copyright holders.

All rights reserved. This program and the accompanying materials
are made available under the terms of the Eclipse Public License v1.0
which accompanies this distribution, and is available at
http://www.eclipse.org/legal/epl-v10.html

Contributors:
  Andre Lengwenus - port to Python and further improvements
  Tobias Juettner - initial LCN binding for openHAB (Java)
"""

import asyncio
from asyncio import Task
from collections import deque
from typing import (
    TYPE_CHECKING,
    Any,
    Awaitable,
    Callable,
    Deque,
    Dict,
    List,
    Optional,
    Sequence,
    Tuple,
    Union,
)

from pypck import inputs, lcn_defs
from pypck.lcn_addr import LcnAddr
from pypck.pck_commands import PckGenerator
from pypck.timeout_retry import DEFAULT_TIMEOUT_MSEC, TimeoutRetryHandler

if TYPE_CHECKING:
    from pypck.connection import PchkConnectionManager


class SerialRequestHandler:
    """Request handler to request serial number information from module."""

    def __init__(
        self,
        addr_conn: "ModuleConnection",
        num_tries: int = 3,
        timeout_msec: int = 1500,
        software_serial: Optional[int] = None,
    ):
        """Initialize class instance."""
        self.addr_conn = addr_conn

        self.hardware_serial = -1
        self.manu = -1
        if software_serial is None:
            software_serial = -1
        self.software_serial = software_serial
        self.hardware_type = -1

        # Serial Number request
        self.trh = TimeoutRetryHandler(num_tries, timeout_msec)
        self.trh.set_timeout_callback(self.timeout)

        # callback
        addr_conn.register_for_inputs(self.process_input)

        # events
        self.serial_known = asyncio.Event()

    def process_input(self, inp: inputs.Input) -> None:
        """Create a task to process the input object concurrently."""
        asyncio.create_task(self.async_process_input(inp))

    async def async_process_input(self, inp: inputs.Input) -> None:
        """Process incoming input object.

        Method to handle incoming commands for this specific request handler.
        """
        if isinstance(inp, inputs.ModSn):
            self.hardware_serial = inp.serial
            self.manu = inp.manu
            self.software_serial = inp.sw_age
            self.hardware_type = inp.hw_type

            if not self.serial_known.is_set():
                self.serial_known.set()

            await self.cancel()

    def timeout(self, failed: bool = False) -> None:
        """Is called on serial request timeout."""
        if not failed:
            self.addr_conn.send_command(False, PckGenerator.request_serial())
        else:
            self.serial_known.set()

    async def request(self) -> Dict[str, int]:
        """Request serial number."""
        await self.addr_conn.conn.segment_scan_completed_event.wait()
        self.serial_known.clear()
        self.trh.activate()
        await self.serial_known.wait()
        return self.serial

    async def cancel(self) -> None:
        """Cancel serial number request."""
        await self.trh.cancel()

    @property
    def serial(self) -> Dict[str, int]:
        """Return serial numbers of a module."""
        return {
            "hardware_serial": self.hardware_serial,
            "manu": self.manu,
            "software_serial": self.software_serial,
            "hardware_type": self.hardware_type,
        }


class NameCommentRequestHandler:
    """Request handler to request name, comment and OEM text of a module."""

    def __init__(
        self,
        addr_conn: "ModuleConnection",
        num_tries: int = 3,
        timeout_msec: int = 1500,
    ):
        """Initialize class instance."""
        self.addr_conn = addr_conn

        self._name: List[Optional[str]] = [None] * 2
        self._comment: List[Optional[str]] = [None] * 3
        self._oem_text: List[Optional[str]] = [None] * 4

        # Name requests
        self.name_trhs = []
        for block_id in range(2):
            trh = TimeoutRetryHandler(num_tries, timeout_msec)
            trh.set_timeout_callback(self.timeout_name, block_id=block_id)
            self.name_trhs.append(trh)

        self.comment_trhs = []
        for block_id in range(3):
            trh = TimeoutRetryHandler(num_tries, timeout_msec)
            trh.set_timeout_callback(self.timeout_comment, block_id=block_id)
            self.comment_trhs.append(trh)

        self.oem_text_trhs = []
        for block_id in range(4):
            trh = TimeoutRetryHandler(num_tries, timeout_msec)
            trh.set_timeout_callback(self.timeout_oem_text, block_id=block_id)
            self.oem_text_trhs.append(trh)

        # callback
        addr_conn.register_for_inputs(self.process_input)

        # events
        self.name_known = asyncio.Event()
        self.comment_known = asyncio.Event()
        self.oem_text_known = asyncio.Event()

    def process_input(self, inp: inputs.Input) -> None:
        """Create a task to process the input object concurrently."""
        asyncio.create_task(self.async_process_input(inp))

    async def async_process_input(self, inp: inputs.Input) -> None:
        """Process incoming input object.

        Method to handle incoming commands for this specific request handler.
        """
        if isinstance(inp, inputs.ModNameComment):
            command = inp.command
            block_id = inp.block_id
            text = inp.text

            if command == "N":
                self._name[block_id] = f"{text:10s}"
                await self.cancel_name(block_id)
                if not self.name_known.is_set() and (None not in self._name):
                    self.name_known.set()
                    await self.cancel_name()

            elif command == "K":
                self._comment[block_id] = f"{text:12s}"
                await self.cancel_comment(block_id)
                if not self.comment_known.is_set() and (None not in self._comment):
                    self.comment_known.set()
                    await self.cancel_comment()

            elif command == "O":
                self._oem_text[block_id] = f"{text:12s}"
                await self.cancel_oem_text(block_id)
                if not self.oem_text_known.is_set() and (None not in self._oem_text):
                    self.oem_text_known.set()
                    await self.cancel_oem_text()

    def timeout_name(self, failed: bool = False, block_id: int = 0) -> None:
        """Is called on serial request timeout."""
        if not failed:
            self.addr_conn.send_command(False, PckGenerator.request_name(block_id))
        else:
            self.name_known.set()

    def timeout_comment(self, failed: bool = False, block_id: int = 0) -> None:
        """Is called on serial request timeout."""
        if not failed:
            self.addr_conn.send_command(False, PckGenerator.request_comment(block_id))
        else:
            self.comment_known.set()

    def timeout_oem_text(self, failed: bool = False, block_id: int = 0) -> None:
        """Is called on serial request timeout."""
        if not failed:
            self.addr_conn.send_command(False, PckGenerator.request_oem_text(block_id))
        else:
            self.oem_text_known.set()

    async def request_name(self) -> str:
        """Request name from a module."""
        self._name = [None] * 2
        await self.addr_conn.conn.segment_scan_completed_event.wait()
        self.name_known.clear()
        for trh in self.name_trhs:
            trh.activate()
        await self.name_known.wait()
        return self.name

    async def request_comment(self) -> str:
        """Request comments from a module."""
        self._comment = [None] * 3
        await self.addr_conn.conn.segment_scan_completed_event.wait()
        self.comment_known.clear()
        for trh in self.comment_trhs:
            trh.activate()
        await self.comment_known.wait()
        return self.comment

    async def request_oem_text(self) -> List[str]:
        """Request OEM text from a module."""
        self._oem_text = [None] * 4
        await self.addr_conn.conn.segment_scan_completed_event.wait()
        self.oem_text_known.clear()
        for trh in self.oem_text_trhs:
            trh.activate()
        await self.oem_text_known.wait()
        return self.oem_text

    async def request(self) -> Tuple[str, str, List[str]]:
        """Request name, comments and OEM text from a module."""
        return await asyncio.gather(
            self.request_name(), self.request_comment(), self.request_oem_text()
        )

    async def cancel_name(self, block_id: Optional[int] = None) -> None:
        """Cancel name request task."""
        if block_id is None:  # cancel all
            for trh in self.name_trhs:
                await trh.cancel()
        else:
            await self.name_trhs[block_id].cancel()

    async def cancel_comment(self, block_id: Optional[int] = None) -> None:
        """Cancel comment request task."""
        if block_id is None:  # cancel all
            for trh in self.comment_trhs:
                await trh.cancel()
        else:
            await self.comment_trhs[block_id].cancel()

    async def cancel_oem_text(self, block_id: Optional[int] = None) -> None:
        """Cancel OEM text request task."""
        if block_id is None:  # cancel all
            for trh in self.oem_text_trhs:
                await trh.cancel()
        else:
            await self.oem_text_trhs[block_id].cancel()

    async def cancel(self) -> None:
        """Cancel all name, comment and OEM text request tasks."""
        await asyncio.gather(
            self.cancel_name(), self.cancel_comment(), self.cancel_oem_text()
        )

    @property
    def name(self) -> str:
        """Return stored name."""
        return "".join([block for block in self._name if block]).strip()

    @property
    def comment(self) -> str:
        """Return stored comment."""
        return "".join([block for block in self._comment if block]).strip()

    @property
    def oem_text(self) -> List[str]:
        """Return stored OEM text."""
        return [block.strip() if block else "" for block in self._oem_text]
        # return {'block{}'.format(idx):text
        #         for idx, text in enumerate(self._oem_text)}

        # return ''.join([block for block in self._oem_text if block])


class ModulePropertiesRequestHandler:
    """Manages all property requestst for serial number, name, comments, ..."""

    def __init__(
        self, addr_conn: "ModuleConnection", software_serial: Optional[int] = None
    ):
        """Construct ModulePropertiesRequestHandler."""
        self.addr_conn = addr_conn
        self.settings = addr_conn.conn.settings

        self.serial_request_task: Optional[Task[Dict[str, int]]] = None

        num_tries: int = self.settings["NUM_TRIES"]
        timeout_msec: int = self.settings["DEFAULT_TIMEOUT_MSEC"]

        # Serial Number request
        self.serials = SerialRequestHandler(
            addr_conn,
            num_tries,
            timeout_msec=timeout_msec,
            software_serial=software_serial,
        )

        # NameComment request
        self.name_comment = NameCommentRequestHandler(
            addr_conn,
            num_tries,
            timeout_msec=timeout_msec,
        )

    async def activate_all(self) -> None:
        """Activate all properties requests."""
        # software_serial is not given externally
        await self.addr_conn.conn.segment_scan_completed_event.wait()
        if self.serials.software_serial == -1:
            self.serial_request_task = asyncio.create_task(self.serials.request())

    async def cancel_all(self) -> None:
        """Cancel all properties requests."""
        if self.serial_request_task is not None:
            self.serial_request_task.cancel()
            try:
                await self.serial_request_task
            except asyncio.CancelledError:
                pass
        await self.serials.cancel()
        await self.name_comment.cancel()


class StatusRequestsHandler:
    """Manages all status requests for variables, software version, ..."""

    def __init__(self, addr_conn: "ModuleConnection"):
        """Construct StatusRequestHandler instance."""
        self.addr_conn = addr_conn
        self.settings = addr_conn.conn.settings

        self.last_requested_var_without_type_in_response = lcn_defs.Var.UNKNOWN
        self.last_var_lock = asyncio.Lock()

        # Output-port request status (0..3)
        self.request_status_outputs = []
        for output_port in range(4):
            trh = TimeoutRetryHandler(
                -1, self.settings["MAX_STATUS_EVENTBASED_VALUEAGE_MSEC"]
            )
            trh.set_timeout_callback(self.request_status_outputs_timeout, output_port)
            self.request_status_outputs.append(trh)

        # Relay request status (all 8)
        self.request_status_relays = TimeoutRetryHandler(
            -1, self.settings["MAX_STATUS_EVENTBASED_VALUEAGE_MSEC"]
        )
        self.request_status_relays.set_timeout_callback(
            self.request_status_relays_timeout
        )

        # Binary-sensors request status (all 8)
        self.request_status_bin_sensors = TimeoutRetryHandler(
            -1, self.settings["MAX_STATUS_EVENTBASED_VALUEAGE_MSEC"]
        )
        self.request_status_bin_sensors.set_timeout_callback(
            self.request_status_bin_sensors_timeout
        )

        # Variables request status.
        # Lazy initialization: Will be filled once the firmware version is
        # known.
        self.request_status_vars = {}
        for var in lcn_defs.Var:
            if var != lcn_defs.Var.UNKNOWN:
                self.request_status_vars[var] = TimeoutRetryHandler(
                    -1, self.settings["MAX_STATUS_EVENTBASED_VALUEAGE_MSEC"]
                )
                self.request_status_vars[var].set_timeout_callback(
                    self.request_status_var_timeout, var=var
                )

        # LEDs and logic-operations request status (all 12+4).
        self.request_status_leds_and_logic_ops = TimeoutRetryHandler(
            -1, self.settings["MAX_STATUS_POLLED_VALUEAGE_MSEC"]
        )
        self.request_status_leds_and_logic_ops.set_timeout_callback(
            self.request_status_leds_and_logic_ops_timeout
        )

        # Key lock-states request status (all tables, A-D).
        self.request_status_locked_keys = TimeoutRetryHandler(
            -1, self.settings["MAX_STATUS_POLLED_VALUEAGE_MSEC"]
        )
        self.request_status_locked_keys.set_timeout_callback(
            self.request_status_locked_keys_timeout
        )

    def preprocess_modstatusvar(self, inp: inputs.ModStatusVar) -> inputs.Input:
        """Fill typeless response with last requested variable type."""
        if inp.orig_var == lcn_defs.Var.UNKNOWN:
            # Response without type (%Msssaaa.wwwww)
            inp.var = self.last_requested_var_without_type_in_response

            self.last_requested_var_without_type_in_response = lcn_defs.Var.UNKNOWN

            if self.last_var_lock.locked():
                self.last_var_lock.release()
        else:
            # Response with variable type (%Msssaaa.Avvvwww)
            inp.var = inp.orig_var

        return inp

    def request_status_outputs_timeout(
        self, failed: bool = False, output_port: int = 0
    ) -> None:
        """Is called on output status request timeout."""
        if not failed:
            self.addr_conn.send_command(
                False, PckGenerator.request_output_status(output_port)
            )

    def request_status_relays_timeout(self, failed: bool = False) -> None:
        """Is called on relay status request timeout."""
        if not failed:
            self.addr_conn.send_command(False, PckGenerator.request_relays_status())

    def request_status_bin_sensors_timeout(self, failed: bool = False) -> None:
        """Is called on binary sensor status request timeout."""
        if not failed:
            self.addr_conn.send_command(
                False, PckGenerator.request_bin_sensors_status()
            )

    async def request_status_var_timeout(
        self, failed: bool = False, var: Optional[lcn_defs.Var] = None
    ) -> None:
        """Is called on variable status request timeout."""
        assert var is not None
        # Detect if we can send immediately or if we have to wait for a
        # "typeless" response first
        has_type_in_response = lcn_defs.Var.has_type_in_response(
            var, self.addr_conn.software_serial
        )
        if not has_type_in_response:
            # Use the chance to remove a failed "typeless variable" request
            try:
                await asyncio.wait_for(self.last_var_lock.acquire(), timeout=3.0)
            except asyncio.TimeoutError:
                pass
            self.last_requested_var_without_type_in_response = var

        # Send variable request
        self.addr_conn.send_command(
            False,
            PckGenerator.request_var_status(var, self.addr_conn.software_serial),
        )

    def request_status_leds_and_logic_ops_timeout(self, failed: bool = False) -> None:
        """Is called on leds/logical ops status request timeout."""
        if not failed:
            self.addr_conn.send_command(
                False, PckGenerator.request_leds_and_logic_ops()
            )

    def request_status_locked_keys_timeout(self, failed: bool = False) -> None:
        """Is called on locked keys status request timeout."""
        if not failed:
            self.addr_conn.send_command(False, PckGenerator.request_key_lock_status())

    async def activate(self, item: Any) -> None:
        """Activate status requests for given item."""
        await self.addr_conn.conn.segment_scan_completed_event.wait()
        # handle variables independently
        if (item in lcn_defs.Var) and (item != lcn_defs.Var.UNKNOWN):
            # wait until we know the software version
            await self.addr_conn.serial_known
            if self.addr_conn.software_serial >= 0x170206:
                timeout_msec = self.settings["MAX_STATUS_EVENTBASED_VALUEAGE_MSEC"]
            else:
                timeout_msec = self.settings["MAX_STATUS_POLLED_VALUEAGE_MSEC"]
            self.request_status_vars[item].set_timeout_msec(timeout_msec)
            self.request_status_vars[item].activate()
        elif item in lcn_defs.OutputPort:
            self.request_status_outputs[item.value].activate()
        elif item in lcn_defs.RelayPort:
            self.request_status_relays.activate()
        elif item in lcn_defs.MotorPort:
            self.request_status_relays.activate()
        elif item in lcn_defs.BinSensorPort:
            self.request_status_bin_sensors.activate()
        elif item in lcn_defs.LedPort:
            self.request_status_leds_and_logic_ops.activate()
        elif item in lcn_defs.Key:
            self.request_status_locked_keys.activate()

    async def cancel(self, item: Any) -> None:
        """Cancel status request for given item."""
        # handle variables independently
        if (item in lcn_defs.Var) and (item != lcn_defs.Var.UNKNOWN):
            await self.request_status_vars[item].cancel()
            self.last_requested_var_without_type_in_response = lcn_defs.Var.UNKNOWN
        elif item in lcn_defs.OutputPort:
            await self.request_status_outputs[item.value].cancel()
        elif item in lcn_defs.RelayPort:
            await self.request_status_relays.cancel()
        elif item in lcn_defs.MotorPort:
            await self.request_status_relays.cancel()
        elif item in lcn_defs.BinSensorPort:
            await self.request_status_bin_sensors.cancel()
        elif item in lcn_defs.LedPort:
            await self.request_status_leds_and_logic_ops.cancel()
        elif item in lcn_defs.Key:
            await self.request_status_locked_keys.cancel()

    async def activate_all(self, activate_s0: bool = False) -> None:
        """Activate all status requests."""
        await self.addr_conn.conn.segment_scan_completed_event.wait()
        for item in (
            list(lcn_defs.OutputPort)
            + list(lcn_defs.RelayPort)
            + list(lcn_defs.BinSensorPort)
            + list(lcn_defs.LedPort)
            + list(lcn_defs.Key)
            + list(lcn_defs.Var)
        ):
            if isinstance(item, lcn_defs.Var) and item == lcn_defs.Var.UNKNOWN:
                continue
            if (
                (not activate_s0)
                and isinstance(item, lcn_defs.Var)
                and (item in lcn_defs.Var.s0s)  # type: ignore
            ):
                continue
            await self.activate(item)

    async def cancel_all(self) -> None:
        """Cancel all status requests."""
        for item in (
            list(lcn_defs.OutputPort)
            + list(lcn_defs.RelayPort)
            + list(lcn_defs.BinSensorPort)
            + list(lcn_defs.LedPort)
            + list(lcn_defs.Key)
            + list(lcn_defs.Var)
        ):
            if isinstance(item, lcn_defs.Var) and item == lcn_defs.Var.UNKNOWN:
                continue
            await self.cancel(item)


class AbstractConnection(LcnAddr):
    """Organizes communication with a specific module.

    Sends status requests to the connection and handles status responses.
    """

    def __init__(
        self,
        conn: "PchkConnectionManager",
        seg_id: int,
        addr_id: int,
        is_group: bool,
        sw_age: Optional[int] = None,
    ):
        """Construct AbstractConnection instance."""
        self.conn = conn
        super().__init__(seg_id=seg_id, addr_id=addr_id, is_group=is_group)

        self._sw_age: Optional[int] = sw_age
        self._serial: Optional[int] = None
        self._manu: Optional[int] = None
        self._hw_type: Optional[int] = None

        self.input_callbacks: List[Callable[[inputs.Input], None]] = []

    def get_sw_age(self) -> Optional[int]:
        """Return standard sw_age."""
        return self._sw_age

    def send_command(self, wants_ack: bool, pck: str) -> None:
        """Send a command to the module represented by this class.

        :param    bool    wants_ack:    Also send a request for acknowledge.
        :param    str     pck:          PCK command (without header).
        """
        self.conn.send_command(
            PckGenerator.generate_address_header(
                self, self.conn.local_seg_id, wants_ack
            )
            + pck
        )

    # ##
    # ## Methods for handling input objects
    # ##

    async def async_process_input(self, inp: inputs.Input) -> None:
        """Is called by input object's process method.

        Method to handle incoming commands for this specific module (status,
        toggle_output, switch_relays, ...)
        """
        for input_callback in self.input_callbacks:
            input_callback(inp)

    def register_for_inputs(
        self, callback: Callable[[inputs.Input], None]
    ) -> Callable[..., None]:
        """Register a function for callback on PCK message received.

        Returns a function to unregister the callback.
        """
        self.input_callbacks.append(callback)
        return lambda callback=callback: self.input_callbacks.remove(callback)

    # ##
    # ## Methods for sending PCK commands
    # ##

    def dim_output(self, output_id: int, percent: float, ramp: int) -> None:
        """Send a dim command for a single output-port.

        :param    int    output_id:    Output id 0..3
        :param    float    percent:      Brightness in percent 0..100
        :param    int    ramp:         Ramp time in milliseconds
        """
        self.send_command(
            not self.is_group(), PckGenerator.dim_ouput(output_id, percent, ramp)
        )

    def dim_all_outputs(self, percent: float, ramp: int, is1805: bool = False) -> None:
        """Send a dim command for all output-ports.

        :param    float    percent:    Brightness in percent 0..100
        :param    int    ramp:       Ramp time in milliseconds.
        :param    bool   is1805:     True if the target module's firmware is
                                     180501 or newer, otherwise False
        """
        self.send_command(
            not self.is_group(), PckGenerator.dim_all_outputs(percent, ramp, is1805)
        )

    def rel_output(self, output_id: int, percent: float) -> None:
        """Send a command to change the value of an output-port.

        :param     int    output_id:    Output id 0..3
        :param     float    percent:      Relative brightness in percent
                                        -100..100
        """
        self.send_command(
            not self.is_group(), PckGenerator.rel_output(output_id, percent)
        )

    def toggle_output(self, output_id: int, ramp: int) -> None:
        """Send a command that toggles a single output-port.

        Toggle mode: (on->off, off->on).

        :param    int    output_id:    Output id 0..3
        :param    int    ramp:         Ramp time in milliseconds
        """
        self.send_command(
            not self.is_group(), PckGenerator.toggle_output(output_id, ramp)
        )

    def toggle_all_outputs(self, ramp: int) -> None:
        """Generate a command that toggles all output-ports.

        Toggle Mode:  (on->off, off->on).

        :param    int    ramp:        Ramp time in milliseconds
        """
        self.send_command(not self.is_group(), PckGenerator.toggle_all_outputs(ramp))

    def control_relays(self, states: List[lcn_defs.RelayStateModifier]) -> None:
        """Send a command to control relays.

        :param    states:   The 8 modifiers for the relay states as alist
        :type     states:   list(:class:`~pypck.lcn_defs.RelayStateModifier`)
        """
        self.send_command(not self.is_group(), PckGenerator.control_relays(states))

    def control_motors_relays(self, states: List[lcn_defs.MotorStateModifier]) -> None:
        """Send a command to control motors via relays.

        :param    states:   The 4 modifiers for the cover states as a list
        :type     states:   list(:class: `~pypck.lcn-defs.MotorStateModifier`)
        """
        self.send_command(
            not self.is_group(), PckGenerator.control_motors_relays(states)
        )

    def control_motors_outputs(
        self,
        state: lcn_defs.MotorStateModifier,
        reverse_time: Optional[lcn_defs.MotorReverseTime] = None,
    ) -> None:
        """Send a command to control a motor via output ports 1+2.

        :param    MotorStateModifier  state: The modifier for the cover state
        :param    MotorReverseTime    reverse_time: Reverse time for modules
                                                    with FW<190C
        :type     state:   :class: `~pypck.lcn-defs.MotorStateModifier`
        """
        self.send_command(
            not self.is_group(),
            PckGenerator.control_motors_outputs(state, reverse_time),
        )

    def activate_scene(
        self,
        register_id: int,
        scene_id: int,
        output_ports: Sequence[lcn_defs.OutputPort] = (),
        relay_ports: Sequence[lcn_defs.RelayPort] = (),
        ramp: Optional[int] = None,
    ) -> None:
        """Activate the stored states for the given scene.

        :param    int                register_id:    Register id 0..9
        :param    int                scene_id:       Scene id 0..9
        :param    list(OutputPort)   output_ports:   Output ports to activate
                                                     as list
        :param    list(RelayPort)    relay_ports:    Relay ports to activate
                                                     as list
        :param    int                ramp:           Ramp value
        """
        self.send_command(
            not self.is_group(), PckGenerator.change_scene_register(register_id)
        )
        if output_ports:
            self.send_command(
                not self.is_group(),
                PckGenerator.activate_scene_output(scene_id, output_ports, ramp),
            )
        if relay_ports:
            self.send_command(
                not self.is_group(),
                PckGenerator.activate_scene_relay(scene_id, relay_ports),
            )

    def store_scene(
        self,
        register_id: int,
        scene_id: int,
        output_ports: Sequence[lcn_defs.OutputPort] = (),
        relay_ports: Sequence[lcn_defs.RelayPort] = (),
        ramp: Optional[int] = None,
    ) -> None:
        """Store states in the given scene.

        :param    int                register_id:    Register id 0..9
        :param    int                scene_id:       Scene id 0..9
        :param    list(OutputPort)   output_ports:   Output ports to store
                                                     as list
        :param    list(RelayPort)    relay_ports:    Relay ports to store
                                                     as list
        :param    int                ramp:           Ramp value
        """
        self.send_command(
            not self.is_group(), PckGenerator.change_scene_register(register_id)
        )
        if output_ports:
            self.send_command(
                not self.is_group(),
                PckGenerator.store_scene_output(scene_id, output_ports, ramp),
            )
        if relay_ports:
            self.send_command(
                not self.is_group(),
                PckGenerator.store_scene_relay(scene_id, relay_ports),
            )

    def var_abs(
        self,
        var: lcn_defs.Var,
        value_or_float: Union[float, lcn_defs.VarValue],
        unit: lcn_defs.VarUnit = lcn_defs.VarUnit.NATIVE,
        is2013: Optional[bool] = None,
    ) -> None:
        """Send a command to set the absolute value to a variable.

        :param     Var        var:      Variable
        :param     float      value:    Absolute value to set
        :param     VarUnit    unit:     Unit of variable
        """
        if isinstance(value_or_float, lcn_defs.VarValue):
            value = value_or_float
        else:
            value = lcn_defs.VarValue.from_var_unit(value_or_float, unit, True)

        if is2013 is not None:
            sw_is2013 = is2013
        elif self._sw_age is not None:
            sw_is2013 = self._sw_age >= 0x170206
        else:
            sw_is2013 = False

        if lcn_defs.Var.to_var_id(var) != -1:
            # Absolute commands for variables 1-12 are not supported
            if self.get_id() == 4 and self.is_group():
                # group 4 are status messages
                self.send_command(
                    not self.is_group(),
                    PckGenerator.update_status_var(var, value.to_native()),
                )
            else:
                # We fake the missing command by using reset and relative
                # commands.
                self.send_command(
                    not self.is_group(), PckGenerator.var_reset(var, sw_is2013)
                )
                self.send_command(
                    not self.is_group(),
                    PckGenerator.var_rel(
                        var, lcn_defs.RelVarRef.CURRENT, value.to_native(), sw_is2013
                    ),
                )
        else:
            self.send_command(
                not self.is_group(), PckGenerator.var_abs(var, value.to_native())
            )

    def var_reset(self, var: lcn_defs.Var, is2013: Optional[bool] = None) -> None:
        """Send a command to reset the variable value.

        :param    Var    var:    Variable
        """
        if is2013 is not None:
            sw_is2013 = is2013
        elif self._sw_age is not None:
            sw_is2013 = self._sw_age >= 0x170206
        else:
            sw_is2013 = False

        self.send_command(not self.is_group(), PckGenerator.var_reset(var, sw_is2013))

    def var_rel(
        self,
        var: lcn_defs.Var,
        value_or_float: Union[float, lcn_defs.VarValue],
        unit: lcn_defs.VarUnit = lcn_defs.VarUnit.NATIVE,
        value_ref: lcn_defs.RelVarRef = lcn_defs.RelVarRef.CURRENT,
        is2013: Optional[bool] = None,
    ) -> None:
        """Send a command to change the value of a variable.

        :param     Var        var:      Variable
        :param     float      value:    Relative value to add (may also be
                                        negative)
        :param     VarUnit    unit:     Unit of variable
        """
        if isinstance(value_or_float, lcn_defs.VarValue):
            value = value_or_float
        else:
            value = lcn_defs.VarValue.from_var_unit(value_or_float, unit, True)

        if is2013 is not None:
            sw_is2013 = is2013
        elif self._sw_age is not None:
            sw_is2013 = self._sw_age >= 0x170206
        else:
            sw_is2013 = False

        self.send_command(
            not self.is_group(),
            PckGenerator.var_rel(var, value_ref, value.to_native(), sw_is2013),
        )

    def lock_regulator(self, reg_id: int, state: bool) -> None:
        """Send a command to lock a regulator.

        :param    int        reg_id:        Regulator id
        :param    bool       state:         Lock state (locked=True,
                                            unlocked=False)
        """
        if reg_id != -1:
            self.send_command(
                not self.is_group(), PckGenerator.lock_regulator(reg_id, state)
            )

    def control_led(self, led: lcn_defs.LedPort, state: lcn_defs.LedStatus) -> None:
        """Send a command to control a led.

        :param    LedPort      led:        Led port
        :param    LedStatus    state:      Led status
        """
        self.send_command(
            not self.is_group(), PckGenerator.control_led(led.value, state)
        )

    def send_keys(self, keys: List[List[bool]], cmd: lcn_defs.SendKeyCommand) -> None:
        """Send a command to send keys.

        :param    list(bool)[4][8]    keys:    2d-list with [table_id][key_id]
                                               bool values, if command should
                                               be sent to specific key
        :param    SendKeyCommand      cmd:     command to send for each table
        """
        for table_id, key_states in enumerate(keys):
            if True in key_states:
                cmds = [lcn_defs.SendKeyCommand.DONTSEND] * 4
                cmds[table_id] = cmd
                self.send_command(
                    not self.is_group(), PckGenerator.send_keys(cmds, key_states)
                )

    def send_keys_hit_deferred(
        self, keys: List[List[bool]], delay_time: int, delay_unit: lcn_defs.TimeUnit
    ) -> None:
        """Send a command to send keys deferred.

        :param    list(bool)[4][8]    keys:          2d-list with
                                                     [table_id][key_id] bool
                                                     values, if command should
                                                     be sent to specific key
        :param    int                 delay_time:    Delay time
        :param    TimeUnit            delay_unit:    Unit of time
        """
        for table_id, key_states in enumerate(keys):
            if True in key_states:
                self.send_command(
                    not self.is_group(),
                    PckGenerator.send_keys_hit_deferred(
                        table_id, delay_time, delay_unit, key_states
                    ),
                )

    def lock_keys(
        self, table_id: int, states: List[lcn_defs.KeyLockStateModifier]
    ) -> None:
        """Send a command to lock keys.

        :param    int                     table_id:  Table id: 0..3
        :param    keyLockStateModifier    states:    The 8 modifiers for the
                                                     key lock states as a list
        """
        self.send_command(not self.is_group(), PckGenerator.lock_keys(table_id, states))

    def lock_keys_tab_a_temporary(
        self, delay_time: int, delay_unit: lcn_defs.TimeUnit, states: List[bool]
    ) -> None:
        """Send a command to lock keys in table A temporary.

        :param    int        delay_time:    Time to lock keys
        :param    TimeUnit   delay_unit:    Unit of time
        :param    list(bool) states:        The 8 lock states of the keys as
                                            list (locked=True, unlocked=False)
        """
        self.send_command(
            not self.is_group(),
            PckGenerator.lock_keys_tab_a_temporary(delay_time, delay_unit, states),
        )

    def dyn_text(self, row_id: int, text: str) -> None:
        """Send dynamic text to a module.

        :param    int    row_id:    Row id 0..3
        :param    str    text:      Text to send (up to 60 bytes)
        """
        encoded_text = text.encode(lcn_defs.LCN_ENCODING)

        parts = [encoded_text[12 * p : 12 * p + 12] for p in range(5)]
        for part_id, part in enumerate(parts):
            if part:
                self.send_command(
                    not self.is_group(),
                    PckGenerator.dyn_text_part(row_id, part_id, part),
                )

    def beep(self, sound: lcn_defs.BeepSound, count: int) -> None:
        """Send a command to make count number of beep sounds.

        :param    BeepSound sound:  Beep sound style
        :param    int       count:  Number of beeps (1..15)
        """
        self.send_command(not self.is_group(), PckGenerator.beep(sound, count))

    def ping(self) -> None:
        """Send a command that does nothing and request an acknowledgement."""
        self.send_command(True, PckGenerator.empty())

    def pck(self, pck: str) -> None:
        """Send arbitrary PCK command.

        :param    str    pck:    PCK command
        """
        self.send_command(not self.is_group(), pck)


class GroupConnection(AbstractConnection):
    """Organizes communication with a specific group.

    It is assumed that all modules within this group are newer than FW170206
    """

    def __init__(
        self,
        conn: "PchkConnectionManager",
        seg_id: int,
        grp_id: int,
        sw_age: int = 0x170206,
    ):
        """Construct GroupConnection instance."""
        super().__init__(conn, seg_id, grp_id, True, sw_age=sw_age)

    def var_abs(
        self,
        var: lcn_defs.Var,
        value: Union[float, lcn_defs.VarValue],
        unit: lcn_defs.VarUnit = lcn_defs.VarUnit.NATIVE,
        is2013: Optional[bool] = None,
    ) -> None:
        """Send a command to set the absolute value to a variable.

        :param     Var        var:      Variable
        :param     float      value:    Absolute value to set
        :param     VarUnit    unit:     Unit of variable
        """
        # for new modules (>=0x170206)
        super().var_abs(var, value, unit, is2013=True)

        # for old modules (<0x170206)
        if var in [
            lcn_defs.Var.TVAR,
            lcn_defs.Var.R1VAR,
            lcn_defs.Var.R2VAR,
            lcn_defs.Var.R1VARSETPOINT,
            lcn_defs.Var.R2VARSETPOINT,
        ]:
            super().var_abs(var, value, unit, is2013=False)

    def var_reset(self, var: lcn_defs.Var, is2013: Optional[bool] = None) -> None:
        """Send a command to reset the variable value.

        :param    Var    var:    Variable
        """
        super().var_reset(var, is2013=True)
        if var in [
            lcn_defs.Var.TVAR,
            lcn_defs.Var.R1VAR,
            lcn_defs.Var.R2VAR,
            lcn_defs.Var.R1VARSETPOINT,
            lcn_defs.Var.R2VARSETPOINT,
        ]:
            super().var_reset(var, is2013=False)

    def var_rel(
        self,
        var: lcn_defs.Var,
        value: Union[float, lcn_defs.VarValue],
        unit: lcn_defs.VarUnit = lcn_defs.VarUnit.NATIVE,
        value_ref: lcn_defs.RelVarRef = lcn_defs.RelVarRef.CURRENT,
        is2013: Optional[bool] = None,
    ) -> None:
        """Send a command to change the value of a variable.

        :param     Var        var:      Variable
        :param     float      value:    Relative value to add (may also be
                                        negative)
        :param     VarUnit    unit:     Unit of variable
        """
        super().var_rel(var, value, is2013=True)
        if var in [
            lcn_defs.Var.TVAR,
            lcn_defs.Var.R1VAR,
            lcn_defs.Var.R2VAR,
            lcn_defs.Var.R1VARSETPOINT,
            lcn_defs.Var.R2VARSETPOINT,
            lcn_defs.Var.THRS1,
            lcn_defs.Var.THRS2,
            lcn_defs.Var.THRS3,
            lcn_defs.Var.THRS4,
            lcn_defs.Var.THRS5,
        ]:
            super().var_rel(var, value, is2013=False)

    async def activate_status_request_handler(self, item: Any) -> None:
        """Activate a specific TimeoutRetryHandler for status requests."""
        await self.conn.segment_scan_completed_event.wait()

    async def activate_status_request_handlers(self) -> None:
        """Activate all TimeoutRetryHandlers for status requests."""
        # self.request_serial.activate()
        await self.conn.segment_scan_completed_event.wait()


class ModuleConnection(AbstractConnection):
    """Organizes communication with a specific module or group."""

    def __init__(
        self,
        conn: "PchkConnectionManager",
        seg_id: int,
        mod_id: int,
        activate_status_requests: bool = False,
        has_s0_enabled: bool = False,
        sw_age: Optional[int] = None,
    ):
        """Construct ModuleConnection instance."""
        super().__init__(conn, seg_id, mod_id, False, sw_age=sw_age)
        self.activate_status_requests = activate_status_requests
        self.has_s0_enabled = has_s0_enabled

        # List of queued PCK commands to be acknowledged by the LCN module.
        # Commands are always without address header.
        # Note that the first one might currently be "in progress".
        self.pck_commands_with_ack: Deque[str] = deque()

        self.request_curr_pck_command_with_ack = TimeoutRetryHandler(
            conn.settings["NUM_TRIES"], conn.settings["DEFAULT_TIMEOUT_MSEC"]
        )
        self.request_curr_pck_command_with_ack.set_timeout_callback(
            self.request_curr_pck_command_with_ack_timeout
        )

        self.properties_requests = ModulePropertiesRequestHandler(
            self, software_serial=sw_age
        )
        self.status_requests = StatusRequestsHandler(self)

        self.activate_prh_task = asyncio.create_task(
            self.activate_properties_request_handlers()
        )
        if self.activate_status_requests:
            self.activate_srh_task = asyncio.create_task(
                self.activate_status_request_handlers()
            )

    def send_command(self, wants_ack: bool, pck: str) -> None:
        """Send a command to the module represented by this class.

        :param    bool    wants_ack:    Also send a request for acknowledge.
        :param    str     pck:          PCK command (without header).
        """
        if wants_ack:
            self.schedule_command_with_ack(pck)
        else:
            super().send_command(False, pck)

    async def activate_properties_request_handlers(self) -> None:
        """Activate all TimeoutRetryHandlers for property requests."""
        await self.properties_requests.activate_all()

    async def activate_status_request_handler(self, item: Any) -> None:
        """Activate a specific TimeoutRetryHandler for status requests."""
        await self.status_requests.activate(item)

    async def activate_status_request_handlers(self) -> None:
        """Activate all TimeoutRetryHandlers for status requests."""
        await self.status_requests.activate_all(activate_s0=self.has_s0_enabled)

    async def cancel_properties_request_handlers(self) -> None:
        """Canecl all TimeoutRetryHandlers for status requests."""
        await self.properties_requests.cancel_all()
        self.activate_prh_task.cancel()
        await self.activate_prh_task

    async def cancel_status_request_handler(self, item: Any) -> None:
        """Cancel a specific TimeoutRetryHandler for status requests."""
        await self.status_requests.cancel(item)
        if self.activate_status_requests:
            self.activate_srh_task.cancel()
            await self.activate_srh_task

    async def cancel_status_request_handlers(self) -> None:
        """Canecl all TimeoutRetryHandlers for status requests."""
        await self.status_requests.cancel_all()

    async def cancel_requests(self) -> None:
        """Cancel all TimeoutRetryHandlers."""
        await self.cancel_status_request_handlers()
        await self.cancel_properties_request_handlers()
        await self.request_curr_pck_command_with_ack.cancel()
        self.pck_commands_with_ack.clear()

    def set_s0_enabled(self, s0_enabled: bool) -> None:
        """Set the activation status for S0 variables.

        :param     bool    s0_enabled:   If True, a BU4L has to be connected
        to the hardware module and S0 mode has to be activated in LCN-PRO.
        """
        self.has_s0_enabled = s0_enabled

    def get_s0_enabled(self) -> bool:
        """Get the activation status for S0 variables."""
        return self.has_s0_enabled

    def get_sw_age(self) -> int:
        """Get the LCN module's firmware date."""
        return self.properties_requests.serials.software_serial

    # ##
    # ## Retry logic if an acknowledge is requested
    # ##

    def schedule_command_with_ack(self, pck: str) -> None:
        """Schedule the next command which requests an acknowledge."""
        # add pck command to pck commands list
        self.pck_commands_with_ack.append(pck)
        # Try to process the new acknowledged command.
        # Will do nothing if another one is still in progress.
        self.try_process_next_command_with_ack()

    async def on_ack(
        self, code: int = -1, timeout_msec: int = DEFAULT_TIMEOUT_MSEC
    ) -> None:
        """Is called whenever an acknowledge is received from the LCN module.

        :param     int    code:           The LCN internal code. -1 means
                                          "positive" acknowledge
        :param     intt   timeout_mSec:   The time to wait for a response
                                          before retrying a request
        """
        # Check if we wait for an ack.
        if self.request_curr_pck_command_with_ack.is_active():
            if self.pck_commands_with_ack:
                self.pck_commands_with_ack.popleft()
            await self.request_curr_pck_command_with_ack.cancel()
            # Try to process next acknowledged command
            self.try_process_next_command_with_ack()

    def try_process_next_command_with_ack(self) -> None:
        """Send the next acknowledged command from the queue."""
        if self.pck_commands_with_ack and (
            not self.request_curr_pck_command_with_ack.is_active()
        ):
            self.request_curr_pck_command_with_ack.activate()

    def request_curr_pck_command_with_ack_timeout(self, failed: bool) -> None:
        """I called on command with acknowledge timeout."""
        # Use the chance to remove a failed command first
        if failed:
            self.pck_commands_with_ack.popleft()
            self.try_process_next_command_with_ack()
        else:
            pck = self.pck_commands_with_ack[0]
            self.conn.send_command(
                PckGenerator.generate_address_header(self, self.conn.local_seg_id, True)
                + pck
            )

    async def async_process_input(self, inp: inputs.Input) -> None:
        """Is called by input object's process method.

        Method to handle incoming commands for this specific module (status,
        toggle_output, switch_relays, ...)
        """
        if isinstance(inp, inputs.ModAck):
            asyncio.create_task(self.on_ack(inp.code, DEFAULT_TIMEOUT_MSEC))
            return None

        # handle typeless variable responses
        if isinstance(inp, inputs.ModStatusVar):
            inp = self.status_requests.preprocess_modstatusvar(inp)

        await super().async_process_input(inp)

    # ##
    # ## Requests
    # ##

    # ## properties

    @property
    def hardware_serial(self) -> int:
        """Return hardware serial of module."""
        return self.properties_requests.serials.hardware_serial

    @property
    def manu(self) -> int:
        """Return manufacturing of module."""
        return self.properties_requests.serials.manu

    @property
    def software_serial(self) -> int:
        """Return software serial of module."""
        return self.properties_requests.serials.software_serial

    @property
    def hw_type(self) -> int:
        """Return hardware type of module."""
        return self.properties_requests.serials.hardware_type

    @property
    def serial(self) -> Tuple[int, int, int, int]:
        """Return serials number information."""
        return (self.hardware_serial, self.manu, self.software_serial, self.hw_type)

    @property
    def name(self) -> str:
        """Retrun stored name."""
        return self.properties_requests.name_comment.name

    @property
    def comment(self) -> str:
        """Return stored comments."""
        return self.properties_requests.name_comment.comment

    @property
    def oem_text(self) -> List[str]:
        """Return stored OEM text."""
        return self.properties_requests.name_comment.oem_text

    # ## future properties

    @property
    def serial_known(self) -> Awaitable[bool]:
        """Check if serials have already been received from module."""
        return self.properties_requests.serials.serial_known.wait()

    async def request_name(self) -> str:
        """Request module name."""
        return await self.properties_requests.name_comment.request_name()

    async def request_comment(self) -> str:
        """Request comments from a module."""
        return await self.properties_requests.name_comment.request_comment()

    async def request_oem_text(self) -> List[str]:
        """Request OEM text from a module."""
        return await self.properties_requests.name_comment.request_oem_text()



"""Test the data flow for Input objects"""

from unittest.mock import patch

import asynctest
import pytest
from pypck.inputs import Input


@pytest.mark.asyncio
async def test_message_to_input(pypck_client):
    """Test data flow from message to input."""
    inp = Input()
    message = "dummy_message"
    pypck_client.async_process_input = asynctest.CoroutineMock()
    with patch("pypck.inputs.InputParser.parse", return_value=[inp]) as inp_parse:
        await pypck_client.process_message(message)

    inp_parse.assert_called_with(message)
    pypck_client.async_process_input.assert_awaited_with(inp)
